/*    ATENCION       ADVERTENCIA       WARNING
Esta utilidad altera un bloque de la flash, no se recomienda su uso
si no se comprende lo que se est haciendo (releer IDblock y Userblock
en el libro). No debera necesitar usarla en los mdulos recientes
Esta versin acompaa esta nota a modo de ejemplo, se recomienda
obtener la ltima versin de la pgina web de Z-World, si necesita usarla
*/


/********************************************************************
	change_userblock.c

	Adds or removes the user block by changing the appropriate
	fields in the ID block on the first flash.

 ********************************************************************

 Instructions:

1.  Define the USERBLOCK macro to 1 to have a user block, or 0 for no
    userblock.

2.  Run the program.  The contents of the ID block will be printed
    (or any errors reported) to stdio.

3.  If you are running Dynamic C 7.06 or later, you will also need
    to edit RABBITBIOS.C to change the MAX_USERBLOCK_SIZE macro.

    - If you are adding a user block, change the value to 0x8000.
    - If you are removing a user block, change the value to 0x1000.

 ********************************************************************/


/********************************************************************/
/**  STEP ONE:  set the HASUSERBLOCK macro to 1 to add a user      **/
/**             block, or to 0 to remove a user block.             **/
/********************************************************************/

#define		HASUSERBLOCK	1

/********************************************************************/
/**  END OF USER-CUSTOMIZABLE SECTION                              **/
/********************************************************************/

// check for several potential problems

#ifndef _FLASH_
#error "This program _must_ be run in flash."
#endif


// function prototypes

int writeIDBlock();
int calcCRC(int initCRC, void *data, int blocksize);

/********************************************************************/

main()
{
	static int	i, crc;
	int	saveMB1;
	char	*div19200ptr;

	//////////////////////////////////////////////////
	// is there a user block?

	if (SysIDBlock.tableVersion == 0x00) {
		printf("This board does not have a user block, which is necessary\n");
		printf("for user block support.  Please run WRITE_IDBLOCK.C to\n");
		printf("create ID and user blocks on this board.\n");
		exit(1);
	}

	//////////////////////////////////////////////////
	// map first two quadrants to flash on /CS0
	// make sure A18 NOT inverted on MB1 (handles bank select)

	saveMB1 = MB1CRShadow;
	WrPortI(MB1CR, &MB1CRShadow, FLASH_WSTATES | 0x00);
	_InitFlashDriver(0x01 | 0x02);

	//////////////////////////////////////////////////
	// reserve space for the user block

	if (HASUSERBLOCK) {
		SysIDBlock.userBlockSize = 0x4000 - (int)SysIDBlock.idBlockSize;
		SysIDBlock.userBlockLoc = 0x4000 - (int)SysIDBlock.idBlockSize;
	} else {
		SysIDBlock.userBlockSize = 0x0000;
		SysIDBlock.userBlockLoc = 0x0000;
	}

	//////////////////////////////////////////////////
	// recalculate the ID block's CRC sum

	SysIDBlock.idBlockCRC = 0x0000;

	i = (int)(&SysIDBlock.reserved) - (int)(&SysIDBlock.tableVersion);
	crc = calcCRC(0x0000, &SysIDBlock, i);
	crc = calcCRC(crc, &SysIDBlock.idBlockSize, 16);

	SysIDBlock.idBlockCRC = crc;

	i = writeIDBlock();
	if (i != 0)	{
		printf("\nError writing ID block (%d): \n", i);
		if (i == -1)	printf("  attempt to write to non-flash area\n");
		else if (i == -2)	printf("  source not located in root\n");
		else if (i == -3)	printf("  timeout during flash write\n");
		else printf("  unknown error type\n");
		exit(1);

	} else
		printf("\nID block successfully written.\n\n");

	//////////////////////////////////////////////////
	// now try and read the new ID block back...

	memset(&SysIDBlock, 0x00, sizeof(SysIDBlock));
	i = _readIDBlock(0x01 | 0x02);

	if (i != 0) {
		printf("Error reading ID block (%d)\n", i);
		exit(1);

	} else {
		printf("ID block read successfully:\n");
		printf("  Version       = %d\n", SysIDBlock.tableVersion);
		printf("  ID block size = %d bytes\n", SysIDBlock.idBlockSize);

		printf("\n");
		printf("  User block offset = %04x\n", SysIDBlock.userBlockLoc);
		printf("  User block size   = %04x\n", SysIDBlock.userBlockSize);
	}

	//////////////////////////////////////////////////
	// restore memory quadrant 1

	WrPortI(MB1CR, &MB1CRShadow, saveMB1);
}

/********************************************************************/

int writeIDBlock()
{
	static unsigned long	physaddr;
	static int	addr, blockSize;
	static char	xpc;
	static int	err;


	blockSize = sizeof(SysIDBlock);
	physaddr = 0x00080000ul - blockSize;

	addr = 0xE000 + (int)(physaddr & 0x0FFFul);
	xpc  = (int)(((physaddr - (unsigned long)addr) & 0xFF000ul) >> 12ul);

	_overwrite_block_flag = 1;
	err = WriteFlash(physaddr, &SysIDBlock, blockSize);
	_overwrite_block_flag = 0;

	return(err);
}

/********************************************************************/

int setTimeStamp()
{
	static struct tm	t;

	tm_rd(&t);
	SysIDBlock.timestamp[0] = (t.tm_year + 1900) / 100;
	SysIDBlock.timestamp[1] = t.tm_year % 100;
	SysIDBlock.timestamp[2] = t.tm_mon;
	SysIDBlock.timestamp[3] = t.tm_mday;
	SysIDBlock.timestamp[4] = t.tm_hour;
	SysIDBlock.timestamp[5] = t.tm_min;
	SysIDBlock.timestamp[6] = t.tm_sec;
}

/********************************************************************/

#define CRC_POLY		0x1021		// 16-bit CRC polynomial
											//		(recommended by CCITT X25 standard)

/*
 *	CRC calculation functions
 *		- can be called from C (calcCRC) or assembly (_calcCRC)
 *		- reference:  "A painless guide to CRC error detection algorithms",
 *							8/19/1993, Ross N. Williams
 */

#asm
//	int calcCRC(int initCRC, void *data, int blocksize);
calcCRC::
	ld		hl, 2
	add	hl, sp
	ld		hl, (hl)
	ex		de, hl				; initial CRC value

	ld		hl, 6					; the "blocksize" variable
	add	hl, sp
	ld		hl, (hl)
	ld		b, h
	ld		c, l					; get data size off stack

	ld		hl, 4					; the "data" variable
	add	hl, sp
	ld		hl, (hl)				; hl now contains pointer to data

;; assembly entry point
_calcCRC::
;		de contains initial CRC value
;		hl contains pointer to data
;		bc contains number of bytes
dataloop:
	push	hl
	ld		h, (hl)				; get next byte into hl (and shift it left 8 bits)
	ld		l, 0x00

	push	bc						; save byte counter
	call	_byteCRC				; call CRC function from BIOS
	pop	bc						; restore byte counter

	pop	hl
	inc	hl
	dec	bc

	xor	a
	cp		b
	jr		nz, dataloop
	cp		c
	jr		nz, dataloop

	ex		de, hl				; CRC returned in hl
	ret
#endasm

//////////////////////////////////////////////////////////////////////

#asm
;;
;;	Calculate the CRC value for a single byte.  Can be called multiple
;;	times (without resetting the CRC value passed to it) to do a CRC
;;	check on a larger block of data.
;;
;;		data byte passed in h (l should be 0x00)
;;		crc value passed, returned in de
;;
_byteCRC::
	push	bc
	push	af

	ld		b, 0x08				; bit counter
crcloop:
	push	bc						; save bit counter
	ld		a, h
	xor	d
	scf
	ccf							; clear carry bit
	rl		de						; roll crc left one bit
	ex		de, hl
	scf
	ccf							; clear carry bit
	rl		de						; roll data left one bit
	ex		de, hl
	bit	7, a					; result from earlier XOR
	jr		z, bit7iszero
bit7isone:
	ld		bc, CRC_POLY		; recommended 16-bit polynomial (X.25)
	ld		a, e
	xor	c
	ld		e, a
	ld		a, d					; XOR crc with polynomial
	xor	b
	ld		d, a
bit7iszero:
	pop	bc						; restore bit counter
	djnz	crcloop

	pop	af
	pop	bc
	ret
#endasm